#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <limits.h>
#include <dirent.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/fs.h>
#include <linux/posix_types.h>
#include <assert.h>
#include <sys/reboot.h>

/* for insmod */
#include <sys/mman.h>
#include <asm/unistd.h>
#include <sys/syscall.h>

#include <apm.h>

#include "play.h"
#include "readfile.h"
#include "getarg.h"
#include "getfdt.h"
#include "splash.h"

static char **cmdv;
static char **args;
static dev_t rootdev;

#define _DEBUG_

#define PARTITION_SEARCHSTRING	"root="
#define PWR_BUTTON_ATTR		"/sys/devices/platform/powerbutton/pb_cold_debounce_state"
#define PWR_BUTTON_ATTR_ACTION	"/sys/devices/platform/powerbutton/pb_action"
#define PWR_BUTTON_ATTR_USB	"/sys/devices/platform/powerbutton/pb_usbstat"
#define DEVNAME			"/dev/dsp"
#define FLIPFLOP_ATTR		"/sys/devices/platform/flipflop/value"
enum {
	FLIPFLOP_UNTOUCHED,
	FLIPFLOP_RESET,
};


/* We configure the oss mixer via the proc interface.
 * at boot time , alsa will take over anyway 
 */
#define PROC_OSS_MIXER_PATH "/proc/asound/card0/oss_mixer"
#define PROC_OSS_MIXER_CONFIGURE_STRING "VOLUME \"SPK\" 0"
#define OSS_MIXER_NUMBER 0 /* the mixer to use (SPK == global volume on bemont) */


#ifdef _DEBUG_
#define VERBOSE
#endif

#ifdef VERBOSE
#define PRINTF(fmt, args...) printf(fmt, ##args)
#else
#define PRINTF(fmt, args...)
#endif

#ifdef _DEBUG_
static void dump_args(int argc, char *argv[])
{
        int i;

        printf("argc == %d\n", argc);

        for (i = 0; i < argc; i++)
	{
                printf("argv[%d]: \"%s\"\n", i, argv[i]);
        }

        if (argv[argc] != NULL)
	{
                printf("argv[%d]: \"%s\" (SHOULD BE NULL)\n", argc, argv[argc]);
        }
}
#endif /* _DEBUG_ */

static int split_cmdline(int cmdcmax, char *cmdv[], char *argv0, char *cmdlines[], char *args[])
{
	int was_space;
	char c, *p;
	int vmax = cmdcmax;
	int v = 1;
	int space;

	if (cmdv)
		cmdv[0] = argv0;

	/* First, add the parsable command lines */
	while (*cmdlines)
	{
		p = *cmdlines++;
		was_space = 1;
		while (v < vmax)
		{
			c = *p;
			space = isspace(c);
			if ((space || !c) && !was_space)
			{
				if (cmdv)
					*p = '\0';
				v++;
			}
			else if (was_space)
				{
				if (cmdv)
					cmdv[v] = p;
				}

			if (!c)
				break;

			was_space = space;
			p++;
		}
	}

	/* Second, add the explicit command line arguments */
	while (*args && v < vmax)
	{
		if (cmdv)
			cmdv[v] = *args;
		v++;
		args++;
	}

	if (cmdv)
		cmdv[v] = NULL;

	return v;
}

static void delete_contents(const char *directory)
{
        DIR *dir;
        struct dirent *d;
        struct stat st;

        /* Don't descend into other filesystems */
        if (lstat(directory, &st) || st.st_dev != rootdev)
		return;

        /* Recursively delete the contents of directories. */
        if (S_ISDIR(st.st_mode))
		{
                dir = opendir(directory);
                if (dir)
		{
                        while ((d = readdir(dir)))
			{
                                char *newdir = d->d_name;

                                /* Skip . and .. */
                                if (*newdir=='.' && (!newdir[1] || (newdir[1]=='.' && !newdir[2])))
                                        continue;

                                /* Recurse to delete contents */
                                newdir = alloca(strlen(directory) + strlen(d->d_name) + 2);
                                sprintf(newdir, "%s/%s", directory, d->d_name);
                                delete_contents(newdir);
                        }
                        closedir(dir);

                        /* Directory should now be empty. Zap it. */
                        rmdir(directory);
                }

        /* It wasn't a directory. Zap it. */
        }
	else
	{
		unlink(directory);
	}
}

/* We use error numbers in a loose translation... */
static const char *moderror(int err)
{
	switch (err) {
		case ENOEXEC:
			return "Invalid module format";
		case ENOENT:
			return "Unknown symbol in module";
		case ESRCH:
			return "Module has wrong symbol version";
		case EINVAL:
			return "Invalid parameters";
		default:
			return strerror(err);
	}
}

static int load_module(const char* filename){
	int fd;
	long int ret;
	struct stat st;
	unsigned long len;
	void *map;

	if(! filename ){
		return -EINVAL;
	}

	if ((fd = open(filename, O_RDONLY, 0)) < 0) {
		int tmp;
		char msg[FILENAME_MAX + 32];

		tmp = errno; /* snprintf might alter errno */
		snprintf( msg, sizeof(msg), "cannot open module `%s'", filename);
		errno = tmp;
		perror( msg );
		return -errno;
	}

	fstat(fd, &st);
	len = st.st_size;
	map = mmap(NULL, len, PROT_READ, MAP_SHARED, fd, 0);
	if (map == MAP_FAILED) {
		printf("cannot mmap `%s'", filename);
		return -1;
	}

	ret = syscall(__NR_init_module, map, len, "");
	if (ret != 0) {
		printf("cannot insert `%s': %s (%li)",
				filename, moderror(errno), ret);
		return ret;
	}

	return 0;
}

static int load_all_modules( const char *dir_name )
{
	DIR *dir;
	struct dirent *d;
	char *newdir;
	int retval=0;

	dir=opendir( dir_name );
	if( dir == NULL )
		return 0;

	while( (d=readdir( dir )) != NULL )
	{
		if( !strcmp( d->d_name, "." ) || !strcmp( d->d_name, ".." ) )
			continue;

		newdir = alloca(strlen(dir_name) + strlen(d->d_name) + 2);
		sprintf(newdir, "%s/%s", dir_name, d->d_name);

		if( load_module( newdir ) )
		{
			retval=-EINVAL;
			printf( "Error loading module %s\n", d->d_name );
		}
	}
	closedir(dir);
	return retval;
}
		
static int remove_module(const char* module_name, unsigned int flags)
{
	int ret;
	flags |=  O_NONBLOCK|O_EXCL;
	ret = syscall(__NR_delete_module, module_name, flags);
	if(ret!= 0){
		printf("Cannot remove module %s: %s (%i)", 
				module_name, moderror(errno), ret);
	}
	return ret;
}

static int mount_sys_fs(const char *check, const char *fsname, const char *fstype)
{
        struct stat st;

        if (stat(check, &st) == 0)
	{
                return 0;
        }

        mkdir(fsname, 0555);

        if (mount("none", fsname, fstype, 0, NULL) == -1)
	{
                return -1;
        }

        return 1;
}

static int get_usbstat(void)
{
	FILE *attr;
	char usbstat[2];
	int result;

	attr = fopen(PWR_BUTTON_ATTR_USB, "r");
	if (attr) {
		result = fread(usbstat, sizeof(usbstat), 1, attr);
		if (result > 0 || (result == 0 && feof(attr))) {
			result = usbstat[0] - '0';
		} else {
			fprintf(stderr, "Error reading from " PWR_BUTTON_ATTR_USB ": %s, result=%d feof=%d\n", strerror(errno), result, feof(attr));
			fflush(stderr);
			result = errno;
		}
		fclose(attr);
	} else {
		fprintf(stderr, "Error opening pb_usb\n");
		result = errno;
	}

	return result;
}

static int get_pwr_button_action(void)
{
	FILE *attr;
	char action_str[16];
	int result;

	attr = fopen(PWR_BUTTON_ATTR_ACTION, "r");
	if (attr)
	{
		result = fread(action_str, 16, 1, attr);

		if (result > 0 || (result == 0 && feof(attr))) {
			if (!strcmp(action_str,"suspend"))
				result = 0;
			else if (!strcmp(action_str, "suicide"))
				result = 1;
			else
				result = -EINVAL;
		} else {
			fprintf(stderr, "Error reading from " PWR_BUTTON_ATTR_ACTION ": %s result=%d\n", strerror(errno), result);
			fflush(stderr);
			result = errno;
		}
		fclose(attr);
	}
	else
	{
		fprintf(stderr, "Error opening pb_action\n");
		result = errno;
	}

	return result;
}

/*
 * Check state of power button
 * Return 0 for off, >0 for on, <0 for error
 */
static int get_pwr_button_state(void)
{
	FILE *attr;
	char button_state;
	int result;

	attr = fopen(PWR_BUTTON_ATTR, "r");
	if (!attr)
	{
		int err = errno;
		/* Carry on if power button support is not enabled in kernel */
		if (err == ENOENT)
		{
			PRINTF("Power button support not enabled\n");
			return 1;
		}
		goto err_open;
	}
	if (!fread((void*) &button_state, sizeof(button_state), 1, attr))
		goto err_read;
	PRINTF("Power button state is [%c]\n", (int) button_state);
	if (button_state == '0')
		result = 0;
	else if (button_state == '1')
		result = 1;
	else 
	{
#ifdef _DEBUG_
		goto err_invalid_state;
#else
		/* If power button support broken, assume it was pressed */
		result = 1;
#endif
	}
	if (fclose(attr))
		goto err_close;
	return result;

	err_open:
		perror("Could not open " PWR_BUTTON_ATTR "\n");
		return -1;
	err_read:
		perror("Could not read file " PWR_BUTTON_ATTR "\n");
		goto err;
	err_invalid_state:
		fprintf(stderr, "Invalid button state detected\n");
	err:
		if (fclose(attr))
			goto err_close;
		return -1;
	err_close:
		perror("Could not close file " PWR_BUTTON_ATTR "\n");
		return -1;
}



int configure_oss_mixer(){
	int fd;
	if ((fd = open (PROC_OSS_MIXER_PATH, O_WRONLY, 0)) == -1){
		perror(PROC_OSS_MIXER_PATH);
		return -1;
	}

	/* write the configuration */
	if (write(fd,PROC_OSS_MIXER_CONFIGURE_STRING,strlen(PROC_OSS_MIXER_CONFIGURE_STRING) + 1) ==-1){
		perror("configure oss mixer");
		close(fd);
		return -1;
	}
	close(fd);
	return 0;
}

void shutdown(void)
{
	PRINTF("*** System is going DOWN ***\n");
	if (!reboot(RB_POWER_OFF))
		while (1) {};
	fprintf(stderr, "Unable to shut system down");
	exit(EXIT_FAILURE);
}

void suspend(void)
{
	int fd;
	/* Suspend */
	fd = open("/sys/power/state", O_RDWR);
	if (fd < 0) {
		printf("Unable to open /sys/power/state.\n\tFailed suspending!\n");
		exit(EXIT_FAILURE);
	}
	write(fd, "mem", strlen("mem"));
	close(fd);
}

void set_printk_level(void)
{
	int fd;

	fd = open("/proc/sys/kernel/printk", O_RDWR);
	if (fd < 0) {
		printf("Failed to set printk levels\n");
		return;
	}

	write(fd, "4 1 1 4", strlen("4 1 1 4"));
	close(fd);
}

void suspend_then_poweroff(int flipflop)
{
 int fd;
 int zero=0;

	suspend();

	if (flipflop) {
		fd = open(FLIPFLOP_ATTR, O_RDWR);
		if (fd < 0) {
			fprintf(stderr, "Failed opening " FLIPFLOP_ATTR " (errno=%d)\n", fd);
			fflush(stderr);
			exit(EXIT_FAILURE);
		}
		write(fd, &zero, 1);
		close(fd);
	}

	/* If we reach this point, it means we resumed: reboot */
	fprintf(stderr, "resuming -> rebooting\n");
	fflush(stderr);
	reboot(RB_POWER_OFF);
}

static int read_apm_batt(unsigned int * battery_percentage, int *online)
{
	int ret;
	apm_info info;
	
	ret = apm_read(&info);
	*battery_percentage = info.battery_percentage;
	*online = info.ac_line_status;

	return ret;
}

int play_in_background(char * pWave)
{
  int fd;
  int audiofd;
  int l;
  unsigned char buf[1024];

  if ((audiofd = open (DEVNAME, O_WRONLY, 0)) == -1) {
    perror (DEVNAME);
    return 1; /* failure */
  }

  if ((fd = open (pWave, O_RDONLY, 0)) == -1) {
    perror (pWave);
    return 1; /* failure */
  }


  if ((l = read (fd, buf, sizeof (buf))) == -1) {
    perror (pWave);
    close (fd);
    close (audiofd);
    return 1; /* failure */
  }

  if (l == 0) {
    fprintf (stderr, "%s is empty file.\n", pWave);
    close (fd);
    close (audiofd);
    return 1 ; /* failure */
  }

  lseek (fd, 0, SEEK_SET);	/* Start from the beginning */

 
  if (fork() == 0){/* this is where we "detatch" */
  	if (l > 16 && memcmp (&buf[0], "RIFF", 4) == 0 && memcmp (&buf[8], "WAVE", 4) == 0) {
	    play_wave (pWave, fd, audiofd , buf, l) ;
	  };
          fprintf(stderr, "AUTOTEST: PLAYING TOMTOM SOUND\n");
          fflush(stderr);
	  close (fd);
	  snd_dsp_sync(audiofd,0);
	  close (audiofd);
	  exit(EXIT_SUCCESS);
  }
  /* also close in the parent */
  close (fd);
  close (audiofd);

  return 0; /* all is just fine */
}

int main(int argc, char **argv) 
{
	int		ret;
	char *		root_dev_name = NULL;
	struct stat	rst;
	struct stat	cst;
	struct stat	ist;
	int		mnt_procfs;
	int		mnt_sysfs;
	int		cmdc;
	int		mountflag;
	int		i;
	char *		cmdlines[3];
	unsigned int	battery_capacity, min_batt_capacity;
	int		online;
	int		pb;
	char		*bootmode=getenv( "sysboot_mode" );

	fprintf(stderr, "init compiled on %s at %s\n", __DATE__, __TIME__);
	fflush(stderr);

	/* Mount /sys */
        mnt_sysfs = mount_sys_fs(PWR_BUTTON_ATTR, "/sys", "sysfs") >= 0;
        if (!mnt_sysfs)
	{
                printf("could not mount sysfs\n");
		exit(EXIT_FAILURE);
        }

	/* Mount /proc */
        mnt_procfs = mount_sys_fs("/proc/cmdline", "/proc", "proc") >= 0;
        if (!mnt_procfs)
	{
                printf("could not mount %s as %s\n", "/proc/cmdline", "/proc");
		return 1;
        }

	/* Set default log level */
	set_printk_level();

	/* Load all present kernel modules. Needed for binary only modules. */
	if( load_all_modules( "/lib/modules" ) != 0 )
		printf( "Error loading kernel modules! Trying to continue\n" );

	if (readfile("/arch.cmd", &cmdlines[0]) < 0)
		cmdlines[0] = "";

	/* Get minimum battery level from DTS */
	min_batt_capacity = fdt_get_ulong("/options/battery",
			"battery-low-suspend", 10);

	/* Get actual battery level. */
	if (read_apm_batt(&battery_capacity, &online) < 0) {
		fprintf(stderr, "Could not read battery level! "
				"Assuming 100\n");
		battery_capacity = 100;
	}

	/* Shutdown if level is too low and not charging. */
	if (!online && (battery_capacity < min_batt_capacity)) {
		fprintf(stderr, "Should shutdown, batt %d < %d (online %d)\n",
			battery_capacity, min_batt_capacity, online);
		//TODO temporarly disabled as battery driver not ready
		//shutdown();
	}

	printf ("BOOTMODE: %s\n", bootmode);

	/* Check the powerbutton only at cold boot. */
	if( (bootmode == NULL) || !strcmp( bootmode, "cold" ) )
	{
		/* Cold boot.... */
		pb = get_pwr_button_state();
		if(pb < 0){
			//exit(EXIT_FAILURE);
			perror("get_pwr_button_state():");
		}

		if (pb == 0){
			switch(get_pwr_button_action()) {
			case 0: /* Suspend */
				fprintf(stderr, "SUSPENDING: PWR BTN DBNCE\n");
				suspend();
				break;
			case 1: /* Suicide */
				if (get_usbstat() == 0) {
					/* Print a message on console for automated testing */
					fprintf(stderr, "AUTOTEST: PWR BTN DBNCE\n");
					fflush(stderr);

					shutdown();
				} else {
					fprintf(stderr, "SUSPENDING+POWERINGOFF: PWR BTN DBNCE\n");
					suspend_then_poweroff(FLIPFLOP_RESET);
				}
				break;
			default:
				fprintf(stderr, "INVALID PWR BTN ACTION %d\n", get_pwr_button_action());
				fflush(stderr);
				exit(EXIT_FAILURE);
			break;
			}
		}

		assert(pb == 1);

		/* Print a message on console for automated testing */
		fprintf(stderr, "AUTOTEST: COLD BOOT\n");
		fflush(stderr);

		
	}
	else
	{
		/* Print a message on console for automated testing */
		fprintf(stderr, "AUTOTEST: WARM BOOT\n");
		fflush(stderr);
	}	


	{
		struct fb_data fb_data;
		ret = open_fb(&fb_data, "/dev/fb0");
		if (ret) {
			printf("Error opening the famebuffer\n");
			goto crap;
		}

		if (fb_data.variable_screen_info.yres * 2 !=
			fb_data.variable_screen_info.yres_virtual) {
			printf("Will not use page flipping\n");
			fb_data.perform_flip = 0;
		} else {
			fb_data.perform_flip = 1;
		}

		if (fb_data.variable_screen_info.bits_per_pixel != 16) {
			printf("Only 16bpp displays are supported\n");
			exit(1);
		}

		printf("Opened the screen (width,height,bbp)=(%i,%i,%i)",
			fb_data.variable_screen_info.xres,
			fb_data.variable_screen_info.yres,
			fb_data.variable_screen_info.bits_per_pixel);

		draw_rle_file(&fb_data, "/logo.rle", IMAGE_ALIGN_CENTER);
		close_fb(&fb_data);

		/* Print a message on console for automated testing */
		fprintf(stderr, "AUTOTEST: SPLASH SCREEN\n");
		fflush(stderr);
	}
crap:
	if (configure_oss_mixer()){
		fprintf(stderr, "Failed to configure the oss mixer\n");
		fflush(stderr);
	}

	/* Set the volume */
	if (set_volume(OSS_MIXER_NUMBER , 80)){
		fprintf(stderr, "Failed to set the volume\n");
		fflush(stderr);
	} 

	/* Start the Play the TomTom sound thread */
	/* play_in_background_but_first_open_the_handles will open the sound.wav and the audio device
	 * before it forks again (2nd time) to the background. The main thread continues and waits
	 * at least untill the audio thread is forked*/
	if (play_in_background("sound.wav") > 0){
		/* Print a message on console for automated testing */
		fprintf(stderr, "AUTOTEST: ERROR PLAYING TOMTOM SOUND\n");
		fflush(stderr);
	}

    	if (readfile("/arch.cmd", &cmdlines[0]) < 0)
                cmdlines[0] = "";

        if (readfile("/proc/cmdline", &cmdlines[1]) < 0)
	{
                printf("cannot read /proc/cmdline\n");
		return 1;
        }

        cmdlines[2] = NULL;

        /* Find an -- argument, and if so append to the command line */
        for (i = 1; i < argc; i++)
	{
                if (!strcmp(argv[i], "--"))
		{
                        i++;
                        break;
                }
        }
        args = &argv[i];        /* Points either to first argument past -- or to the final NULL */

        /* Count the number of arguments */
        cmdc = split_cmdline(INT_MAX, NULL, argv[0], cmdlines, args);

        /* Actually generate the cmdline array */
        cmdv = (char **)alloca((cmdc+1)*sizeof(char *));
        if (split_cmdline(cmdc, cmdv, argv[0], cmdlines, args) != cmdc)
	{
                printf("cannot split cmdline\n");
		return 1;
        }

#ifdef _DEBUG_
        dump_args(cmdc, cmdv);
#endif /* _DEBUG_ */

	/* Get the boot partition from kernel command line */
	if ((root_dev_name = get_arg(cmdc, cmdv, "root=")) == NULL)
	{
		printf("Search of boot partition failed\n");
		return 1;
	}

	mountflag =  MS_NOATIME | MS_NODIRATIME;
	if (get_flag(cmdc, cmdv, "ro"))
	{
		mountflag |= MS_RDONLY;
	}

	for(;;) 
  	{
		if(mount(root_dev_name, "/mnt/rootfs", "ext2", mountflag, NULL)>=0) 
    		{
			break;
		}
		else 
    		{
			printf("Waiting for the root file system(%s)...\n",root_dev_name);
			sleep(1);
		}
	}

	if(chdir("/mnt/rootfs") != 0) 
  	{
		perror("can't cd to /mnt/rootfs root");
		return 1;
	}

	/* Make sure the current directory is not on the same filesystem as the root directory */
	if (lstat("/", &rst) || lstat(".", &cst))
	{
		printf("stat error\n");
		return 1;
	}
	rootdev = rst.st_dev;

	if (rst.st_dev == cst.st_dev)
	{
		printf("current directory on the same filesystem as the root\n");
		return 1;
	}

	/* The rootfs should have /init */
	if (stat("/mnt/rootfs/init", &ist) || !S_ISREG(ist.st_mode))
	{
		printf ("can't find /init on rootfs\n");
		return 1;
	}

	delete_contents("/");

	if (mount(".", "/", NULL, MS_MOVE, NULL) < 0)	
	{
		perror("can't move mount");
		return 1;
	}

	chroot(".");
	chdir("/");

	execl("/sbin/init", "init", NULL);

	perror("exec failed");
	return 1;
}
